import { type BaseRoot, type DateLiteral, type Default, type distillOut, type of, type UndeclaredKeyBehavior, type writeInvalidPropertyKeyMessage } from "@arktype/schema";
import { type anyOrNever, type BigintLiteral, type Dict, type ErrorMessage, type Key, type keyError, type merge, type NumberLiteral, type show } from "@arktype/util";
import type { ParseContext } from "../scope.js";
import type { inferDefinition, validateDefinition } from "./definition.js";
import type { astToString } from "./semantic/utils.js";
import type { validateString } from "./semantic/validate.js";
import type { StringLiteral } from "./string/shift/operand/enclosed.js";
import { Scanner } from "./string/shift/scanner.js";
import type { inferString } from "./string/string.js";
export declare const parseObjectLiteral: (def: Dict, ctx: ParseContext) => BaseRoot;
export declare const writeInvalidUndeclaredBehaviorMessage: (actual: unknown) => string;
export declare const nonLeadingSpreadError = "Spread operator may only be used as the first key in an object";
export type inferObjectLiteral<def extends object, $, args> = show<"..." extends keyof def ? merge<inferDefinition<def["..."], $, args>, _inferObjectLiteral<def, $, args>> : _inferObjectLiteral<def, $, args>>;
/**
 * Infers the contents of an object literal, ignoring a spread definition
 */
type _inferObjectLiteral<def extends object, $, args> = {
    -readonly [k in keyof def as nonOptionalKeyFrom<k, $, args>]: def[k] extends (DefaultValueTuple<infer baseDef, infer defaultValue>) ? def[k] extends anyOrNever ? def[k] : (In?: inferDefinition<baseDef, $, args>) => Default<defaultValue> : def[k] extends DefaultValueString<infer baseDef, infer defaultDef> ? (In?: inferDefinition<baseDef, $, args>) => Default<inferDefinition<defaultDef, $, args>> : inferDefinition<def[k], $, args>;
} & {
    -readonly [k in keyof def as optionalKeyFrom<k>]?: inferDefinition<def[k], $, args>;
};
export type validateObjectLiteral<def, $, args> = {
    [k in keyof def]: k extends IndexKey<infer indexDef> ? validateString<indexDef, $, args> extends ErrorMessage<infer message> ? keyError<message> : inferDefinition<indexDef, $, args> extends (PropertyKey | of<PropertyKey, {}>) ? validateDefinition<def[k], $, args> : keyError<writeInvalidPropertyKeyMessage<indexDef>> : k extends "..." ? inferDefinition<def[k], $, args> extends object ? validateDefinition<def[k], $, args> : keyError<writeInvalidSpreadTypeMessage<astToString<def[k]>>> : k extends "+" ? UndeclaredKeyBehavior : validateDefaultableValue<def, k, $, args>;
};
type validateDefaultableValue<def, k extends keyof def, $, args> = def[k] extends DefaultValueTuple ? validateDefaultValueTuple<def[k], k, $, args> : def[k] extends DefaultValueString ? validateDefaultValueString<def[k], k, $, args> : validateDefinition<def[k], $, args>;
type DefaultValueTuple<baseDef = unknown, defaultValue = unknown> = readonly [
    baseDef,
    "=",
    defaultValue
];
type validateDefaultValueTuple<def extends DefaultValueTuple, k extends PropertyKey, $, args> = parseKey<k>["kind"] extends "required" ? readonly [
    validateDefinition<def[0], $, args>,
    "=",
    inferDefinition<def[0], $, args>
] : ErrorMessage<invalidDefaultKeyKindMessage>;
type DefaultValueString<baseDef extends string = string, defaultDef extends UnitLiteral = UnitLiteral> = `${baseDef}${typeof defaultValueStringOperator}${defaultDef}`;
declare const defaultValueStringOperator = " = ";
type validateDefaultValueString<def extends DefaultValueString, k extends PropertyKey, $, args> = def extends DefaultValueString<infer baseDef, infer defaultDef> ? parseKey<k>["kind"] extends "required" ? validateDefinition<baseDef, $, args> extends (infer e extends ErrorMessage) ? e : [
    distillOut<inferString<baseDef, {}, args>>,
    distillOut<inferString<defaultDef, {}, args>>
] extends [infer base, infer defaultValue] ? defaultValue extends base ? def : ErrorMessage<`${defaultDef} is not assignable to ${baseDef}`> : never : ErrorMessage<invalidDefaultKeyKindMessage> : never;
type nonOptionalKeyFrom<k, $, args> = parseKey<k> extends PreparsedKey<"required", infer inner> ? inner : parseKey<k> extends PreparsedKey<"index", infer inner> ? Extract<inferDefinition<inner, $, args>, Key> : never;
type optionalKeyFrom<k> = parseKey<k> extends PreparsedKey<"optional", infer inner> ? inner : never;
type PreparsedKey<kind extends ParsedKeyKind = ParsedKeyKind, inner extends Key = Key> = {
    kind: kind;
    key: inner;
};
declare namespace PreparsedKey {
    type from<t extends PreparsedKey> = t;
}
type ParsedKeyKind = "required" | "optional" | "index" | MetaKey;
export type MetaKey = "..." | "+";
export type IndexKey<def extends string = string> = `[${def}]`;
interface PreparsedEntry extends PreparsedKey {
    value: unknown;
    default: unknown;
}
declare const unitLiteralKeywords: {
    readonly null: null;
    readonly undefined: undefined;
    readonly true: true;
    readonly false: false;
};
type UnitLiteralKeyword = keyof typeof unitLiteralKeywords;
export type UnitLiteral = StringLiteral | BigintLiteral | NumberLiteral | DateLiteral | UnitLiteralKeyword;
export declare const parseEntry: (key: Key, value: unknown) => PreparsedEntry;
export declare const invalidDefaultKeyKindMessage = "Only required keys may specify default values, e.g. { value: 'number = 0' }";
export type invalidDefaultKeyKindMessage = typeof invalidDefaultKeyKindMessage;
declare const parseKey: (key: Key) => PreparsedKey;
type parseKey<k> = k extends `${infer inner}?` ? inner extends `${infer baseName}${Scanner.EscapeToken}` ? PreparsedKey.from<{
    kind: "required";
    key: `${baseName}?`;
}> : PreparsedKey.from<{
    kind: "optional";
    key: inner;
}> : k extends MetaKey ? PreparsedKey.from<{
    kind: k;
    key: k;
}> : k extends `${Scanner.EscapeToken}${infer escapedMeta extends MetaKey}` ? PreparsedKey.from<{
    kind: "required";
    key: escapedMeta;
}> : k extends IndexKey<infer def> ? PreparsedKey.from<{
    kind: "index";
    key: def;
}> : PreparsedKey.from<{
    kind: "required";
    key: k extends (`${Scanner.EscapeToken}${infer escapedIndexKey extends IndexKey}`) ? escapedIndexKey : k extends Key ? k : `${k & number}`;
}>;
export declare const writeInvalidSpreadTypeMessage: <def extends string>(def: def) => writeInvalidSpreadTypeMessage<def>;
type writeInvalidSpreadTypeMessage<def extends string> = `Spread operand must resolve to an object literal type (was ${def})`;
export {};
